#!/usr/bin/python
#
# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2007,2008, William Trevor King
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.

"""
Separate the more general vib_analyze() from the other vib_*()
functions in calibcant.  Also provide a command line interface
for analyzing data acquired through other workflows.

The relevent physical quantities are :
 Vphoto   The photodiode vertical deflection voltage (what we measure)
"""

import os, time, numpy
import GnuplotBiDir  # used for fitting lorentzian
import common # common module for the calibcant package
import config # config module for the calibcant package
import data_logger
import FFT_tools

def vib_analyze_naive(deflection_bits, Vphoto_in2V=config.Vphoto_in2V,
                      zeroVphoto_bits=config.zeroVphoto_bits) :
    """
    Calculate the variance of the raw data, and convert it to Volts**2.
    This method is simple and easy to understand, but it highly succeptible
    to noise, drift, etc.
    
    Vphoto_in2V is a function converting Vphoto values from bits to Volts.
    This function is assumed linear, see bump_analyze().
    zeroVphoto_bits is the bit value corresponding to Vphoto = zero Volts.
    """
    Vphoto_std_bits = deflection_bits.std()
    Vphoto_std = Vphoto_in2V(Vphoto_std_bits + zpiezo_zeroV_in)
    return Vphoto_std**2
    
def vib_analyze(deflection_bits, freq, minFreq=500, maxFreq=7000,
                chunk_size=4096, overlap=True, window=FFT_tools.window_hann,
                Vphoto_in2V=config.Vphoto_in2V, plotVerbose=False) :
    """
    Calculate the variance in the raw data due to the cantilever's 
    thermal oscillation and convert it to Volts**2.

    Improves on vib_analyze_naive() by first converting Vphoto(t) to 
    frequency space, and fitting a Lorentzian in the relevant frequency
    range (see cantilever_calib.pdf for derivation).

    The conversion to frequency space generates an average power spectrum
    by breaking the data into windowed chunks and averaging the power
    spectrums for the chunks together.
    See FFT_tools.avg_power_spectrum
    
    Vphoto_in2V is a function converting Vphoto values from bits to Volts.
    This function is assumed linear, see bump_analyze().
    zeroVphoto_bits is the bit value corresponding to Vphoto = zero Volts.
    """
    Vphoto_t = numpy.zeros((len(deflection_bits),),
                           dtype=numpy.float)
    # convert the data from bits to volts
    if config.TEXT_VERBOSE : 
        print "Converting %d bit values to voltages" % len(Vphoto_t)
    for i in range(len(deflection_bits)) :
        Vphoto_t[i] = Vphoto_in2V(deflection_bits[i])

    # Compute the average power spectral density per unit time (in uV**2/Hz)
    if config.TEXT_VERBOSE : 
        print "Compute the averaged power spectral density in uV**2/Hz"
    (freq_axis, power) = \
        FFT_tools.unitary_avg_power_spectrum((Vphoto_t - Vphoto_t.mean())*1e6,
                                             freq, chunk_size,overlap, window)

    A,B,C = fit_psd(freq_axis, power, minFreq, maxFreq)

    if config.TEXT_VERBOSE : 
        print "Fitted f(x) = C / ((A**2-x**2)**2 + (x*B)**2) with"
        print "A = %g, \t B = %g, \t C = %g" % (A, B, C)

    vib_plot(deflection_bits, freq_axis, power, A, B, C, plotVerbose=plotVerbose)

    # Our A is in uV**2, so convert back to Volts
    return lorentzian_area(A,B,C) * 1e-12

def fit_psd(freq_axis, psd_data, minFreq=500, maxFreq=7000) :
    """
    freq_axis : array of frequencies in Hz
    psd_data  : array of PSD amplitudes for the frequencies in freq_axis

    Calculate the variance in the raw data due to the cantilever's 
    thermal oscillation and convert it to Volts**2.

    Improves on vib_analyze_naive() by working on frequency space data 
    and fitting a Lorentzian in the relevant frequency range (see
    cantilever_calib.pdf for derivation).

    The conversion to frequency space generates an average power spectrum
    by breaking the data into windowed chunks and averaging the power
    spectrums for the chunks together.
    See FFT_tools.unitary_avg_power_spectrum().    
    """
    # cut out the relevent frequency range
    if config.TEXT_VERBOSE : 
        print "Cut the frequency range %g to %g Hz" % (minFreq, maxFreq)
    imin = 0
    while freq_axis[imin] < minFreq : imin += 1
    imax = imin
    while freq_axis[imax] < maxFreq : imax += 1
    assert imax >= imin + 10 , \
        "Less than 10 points in freq range (%g,%g)" % (minFreq, maxFreq)

    # save about-to-be-fitted data to a temporary file
    datafilename = "%s_%d" % (config.GNUFIT_DATA_BASE, time.time())
    fd = file(datafilename, 'w')
    for i in range(imin, imax) :
        fd.write("%g\t%g\n" % (freq_axis[i], psd_data[i]))
    fd.close()

    # generate guesses for Lorentzian parameters A,B,C
    max_psd_index = numpy.argmax(psd_data[imin:imax]) + imin
    max_psd = psd_data[max_psd_index]
    half_max_index = imin
    while psd_data[half_max_index] < max_psd/2 :
        half_max_index += 1
    res_freq = freq_axis[max_psd_index]
    half_width = freq_axis[max_psd_index] - freq_axis[half_max_index]
    # Lorentzian L(x) = C / ((A**2-x**2)**2 + (B*x)**2)
    # is expected power spectrum for
    # x'' + B x' + A^2 x'' = F_external(t)/m
    # (A = omega_0)
    # C = (2 k_B T B) / (pi m)
    # 
    # A = resonant frequency
    # peak at  x_res = sqrt(A^2 - B^2/2)
    #  which could be complex if there isn't a peak (overdamped)
    # peak height    = C / (3 x_res^4 + A^4)
    # Q = A/B
    #
    # guess Q = 5, and adjust from there
    Q_guess = 5
    # so x_res^2 = B^2 Q^2 - B^2/2 = (Q^2-1/2)B^2 
    #    B = x_res / sqrt(Q^2-1/2)
    if config.TEXT_VERBOSE : 
        print "params : %g, %g" % (res_freq, max_psd)
    B_guess = res_freq / numpy.sqrt(Q_guess**2-0.5)
    A_guess = Q_guess*B_guess
    C_guess = max_psd * (-res_freq**4 + A_guess**4)
    # 
    # Half width w on lower side when L(a-w) = L(a)/2
    #  (a**2 - (a-w)**2)**2 + (b*(a-w))**2 = 2*(b*a)**2
    # Let W=(a-w)**2, A=a**2, and B=b**2
    #  (A - W)**2 + BW = 2*AB
    #  W**2 - 2AW + A**2 + BW = 2AB
    #  W**2 + (B-2A)W + (A**2-2AB) = 0
    #  W = (2A-B)/2 * [1 +/- sqrt(1 - 4(A**2-2AB)/(B-2A)**2]
    #    = (2A-B)/2 * [1 +/- sqrt(1 - 4A(A-2B)/(B-2A)**2]
    #  (a-w)**2 = (2A-B)/2 * [1 +/- sqrt(1 - 4A(A-2B)/(B-2A)**2]
    #  so w is a disaster ;)
    # For some values of A and B (non-underdamped), W is imaginary or negative.
    if config.TEXT_VERBOSE : 
        print "guesses : %g, %g, %g" % (A_guess, B_guess, C_guess)

    # fit Lorentzian using Gnuplot's 'fit' command
    g = GnuplotBiDir.Gnuplot()
    # The Lorentzian is the predicted one-sided power spectral density
    # For an oscillator whose position x obeys
    # m x'' + gamma x' + kx = F_thermal(t)
    #  A is the ?driving noise?
    #  B is gamma, the damping term
    #  C is omega_0, the resonant frequency
    # Fitting the PSD of V = photoSensitivity*x just rescales the parameters
    g("f(x) = C / ((A**2-x**2)**2 + (B*x)**2)")
    ## The mass m is given by m = k_B T / (pi A)
    ## The spring constant k is given by k = m (omega_0)**2
    ## The quality factor Q is given by Q = omega_0 m / gamma
    g("A=%g; B=%g; C=%g" % (A_guess, B_guess, C_guess))
    g("fit f(x) '%s' via A,B,C" % datafilename)
    A=abs(float(g.getvar('A'))) # A and B only show up as squares in f(x)
    B=abs(float(g.getvar('B'))) # so ensure we get positive values
    C=float(g.getvar('C'))
    os.remove(datafilename)
    return (A, B, C)

def lorentzian_area(A, B, C) :
    # Integrating the the power spectral density per unit time (power) over the
    # frequency axis [0, fN] returns the total signal power per unit time
    #  int_0^fN power(f)df = 1/T int_0^T |x(t)**2| dt
    #                      = <V(t)**2>, the variance for our AC signal.
    # The variance from our fitted Lorentzian is the area under the Lorentzian
    #  <V(t)**2> = (pi*C) / (2*B*A**2)
    return (numpy.pi * C) / (2 * B * A**2)

def gnuplot_check_fit(datafilename, A, B, C) :
    """
    return a string containing a gnuplot script displaying the fit.
    """
    string = 'f(x) = C / ((A**2-x**2)**2 + (x*B)**2)\n'
    string += 'A= %g ; B= %g ; C= %g\n' % (A, B, C)
    string += 'set logscale y\n'
    string += "plot '%s', f(x)\n" % datafilename
    return string

def vib_save(data, log_dir=None) :
    """Save the dictionary data, using data_logger.data_log()
    data should be a dictionary of arrays with the fields
      'Z piezo output'
      'Z piezo input'
      'Deflection input'
    """
    if log_dir :
        log = data_logger.data_log(log_dir, noclobber_logsubdir=False,
                                   log_name="vib_%gHz" % freq)
        log.write_dict_of_arrays(data)

def vib_load(datafile=None) :
    """Load the dictionary data, using data_logger.date_load()"
    data should be a dictionary of arrays with the fields
      'Z piezo output'
      'Z piezo input'
      'Deflection input'
    """
    dl = data_logger.data_load()
    data = dl.read_dict_of_arrays(datafile)
    return data

def vib_plot(deflection_bits, freq_axis, power, A, B, C,
             plotVerbose=True) :
    """
    If plotVerbose or config.PYLAB_VERBOSE == True, plot 3 subfigures:
     Time series (Vphoto vs sample index) (show any major drift events),
     A histogram of Vphoto, with a gaussian fit (demonstrate randomness), and
     FFTed Vphoto data (Vphoto vs Frequency) (show major noise components).
    """
    if plotVerbose or config.PYLAB_VERBOSE :
        common._import_pylab()
        common._pylab.hold(False)
        common._pylab.figure(BASE_FIGNUM+2)

        # plot time series
        common._pylab.subplot(311)
        common._pylab.plot(data["Deflection input"], 'r.')
        common._pylab.title("free oscillation")

        # plot histogram distribution and gaussian fit
        common._pylab.subplot(312)
        n, bins, patches = \
            common._pylab.hist(data["Deflection input"], bins=30,
                        normed=1, align='center')
        gaus = numpy.zeros((len(bins),))
        mean = data["Deflection input"].mean()
        std = data["Deflection input"].std()
        pi = numpy.pi
        exp = numpy.exp
        for i in range(len(bins)) :
            gaus[i] = (2*pi)**(-.5)/std * exp(-0.5*((bins[i]-mean)/std)**2)
        common._pylab.hold(True)
        common._pylab.plot(bins, gaus, 'r-');
        common._pylab.hold(False)

        # plot FFTed data
        common._pylab.subplot(313)
        common._pylab.semilogy(freq_axis, power, 'r.-')
        common._flush_plot()
    if (plotVerbose or config.GNUPLOT_VERBOSE) and False : # TODO, clean up and double check...
        # write all the ft data now
        fd = file(datafilename, 'w')
        for i in range(len(freq_axis)) :
            fd.write("%g\t%g\n" % (freq_axis[i], power[i]))
        fd.write("\n") # blank line to separate fit data.
        # write the fit data
        for i in range(imin, imax) :
            x = freq_axis[i]
            fd.write("%g\t%g\n" % (freq_axis[i],
                                   C / ((A**2-x**2)**2 + (x*B)**2) ) )
        fd.close()
        print Vphoto_t.var(), A, B, C, ((numpy.pi * C)/(2 * B * A**2) * 1e-12)
        g("set terminal epslatex color solid")
        g("set output 'calibration.tex'")
        g("set size 2,2") # multipliers on default 5"x3.5"
        #g("set title 'Thermal calibration'")
        g("set logscale y")
        g("set xlabel 'Frequency (Hz)'")
        g("set ylabel 'Power ($\mu$V$^2$/Hz)'")
        g("set xrange [0:20000]")
        g("g(x) = x < %g || x > %g ? 1/0 : f(x)" % (minFreq, maxFreq)) # only plot on fitted region
        g("plot '%s' title 'Vibration data', g(x) title 'Lorentzian fit'" % datafilename)

        g("set mouse")
        g("pause mouse 'click with mouse'")
        g.getvar("MOUSE_BUTTON")
        del(g)

def vib_load_analyze_tweaked(tweak_file, minFreq=1000, maxFreq=7000,
                             chunk_size=2048, overlap=True,
                             window=FFT_tools.window_hann,
                             Vphoto_in2V=config.Vphoto_in2V,
                             plotVerbose=False) :
    """
    Read the vib files listed in tweak_file.
    Return an array of Vphoto variances (due to the cantilever) in Volts**2
    """
    dl = data_logger.data_load()
    Vphoto_var = []
    for path in file(tweak_file, 'r') :
        if path[-1] == '\n':
            path = path.split('\n')[0]
        # read the data
        data = vib_load(path)
        deflection_bits = data['Deflection input']
        freq = float(path.split('_')[-1].split('Hz')[0])
        if config.TEXT_VERBOSE :
            print "Analyzing '%s' at %g Hz" % (path, freq)
        # analyze
        Vphoto_var.append(vib_analyze(deflection_bits, freq, minFreq, maxFreq,
                                      chunk_size, overlap, window,
                                      Vphoto_in2V, plotVerbose))
    return numpy.array(Vphoto_var, dtype=numpy.float)

# commandline interface functions
import scipy.io, sys

def read_data(ifile):
    """
    ifile can be a filename string or open (seekable) file object.
    returns an array of data values (1 column)
    """
    #returns (column 1 array, column 2 array)
    #"""
    if ifile == None :  ifile = sys.stdin
    unlabeled_data=scipy.io.read_array(ifile)
    return unlabeled_data #(unlabeled_data[:,0], unlabeled_data[:,1])

if __name__ == '__main__' :
    # command line interface
    from optparse import OptionParser
    
    usage_string = ('%prog <input-file>\n'
                    '2008, W. Trevor King.\n'
                    '\n'
                    'Deflection power spectral density (PSD) data (X^2/Hz)\n'
                    'and returns the variance in X units (X^2)'
                    '<input-file> should be whitespace-delimited, 2 column ASCII\n'
                    'without a header line.\n'
                    'e.g: "<frequency_Hz>\\t<deflection_psd X^2/Hz>\\n"\n')
    parser = OptionParser(usage=usage_string, version='%prog 0.1')
    parser.add_option('-f', '--sample-freq', dest='freq', default=100e3,
                      help='sample frequency in Hz (default %default)',
                      type='float', metavar='FREQ')
    parser.add_option('-m', '--min-freq', dest='min_freq', default=500,
                      help='minimum frequency in Hz (default %default)',
                      type='float', metavar='FREQ')
    parser.add_option('-M', '--max-freq', dest='max_freq', default=7000,
                      help='maximum frequency in Hz (default %default)',
                      type='float', metavar='FREQ')
    parser.add_option('-o', '--output-file', dest='ofilename',
                      help='write output to FILE (default stdout)',
                      type='string', metavar='FILE')
    parser.add_option('-c', '--comma-out', dest='comma_out', action='store_true',
                      help='Output comma-seperated values (default %default)',
                      default=False)
    parser.add_option('-g', '--gnuplot', dest='gnuplot', action='store_true',
                      help='Print gnuplot fit check script to stderr',
                      default=False)
    parser.add_option('-t', '--tweak-mode', dest='tweakmode', action='store_true',
                      help='Run in tweak-file mode',
                      default=False)
    parser.add_option('-v', '--verbose', dest='verbose', action='store_true',
                      help='Print lots of debugging information',
                      default=False)

    options,args = parser.parse_args()
    parser.destroy()
    assert len(args) >= 1, "Need an input file"
        
    ifilename = args[0]

    if options.ofilename != None :
        ofile = file(options.ofilename, 'w')
    else :
        ofile = sys.stdout
    if options.verbose == True :
        vfile = sys.stderr
    else :
        vfile = None
    config.TEXT_VERBOSE = options.verbose
    config.PYLAB_VERBOSE = False
    config.GNUPLOT_VERBOSE = options.gnuplot

    if options.tweakmode == False :
        data = read_data(ifilename)
        deflection_bits = data['Deflection input']
        Vphoto_var = vib_analyze(deflection_bits, freq=options.freq,
                                 minFreq=options.min_freq,
                                 maxFreq=options.max_freq)
        print >> ofile, Vphoto_var
    else : # tweak mode
        Vphoto_vars = vib_load_analyze_tweaked(ifilename,
                                               minFreq=options.min_freq,
                                               maxFreq=options.max_freq)
        if options.comma_out :
            sep = ','
        else :
            sep = '\n'
        common.write_array(ofile, Vphoto_vars, sep)

    if options.ofilename != None :
        ofile.close()
